# -*- coding: utf-8 -*-
"""
"""
import biosteam as bst
import thermosteam as tmo
import numpy as np
from numpy.testing import assert_allclose

# bst.MultiStageEquilibrium.optimize_result = False

def test_trivial_lle_case():
    import biosteam as bst
    import numpy as np
    from numpy.testing import assert_allclose
    # lle
    with bst.System(algorithm='phenomena based') as sys:
        bst.settings.set_thermo(['Water', 'Octanol', 'Methanol'], cache=True)
        feed = bst.Stream(Water=50, Methanol= 5, units='kg/hr')
        solvent = bst.Stream(Octanol=50, units='kg/hr')
        stage = bst.StageEquilibrium(
            phases=('L', 'l'), ins=[feed, solvent], top_chemical='Octanol',
        )
    sys.simulate()
    streams = stage.ins + stage.outs
    actuals = [i.mol.copy() for i in streams]
    sys.run_phenomena()
    values = [i.mol for i in streams]
    for actual, value in zip(actuals, values):
        assert_allclose(actual, value)

def test_trivial_vle_case():
    import biosteam as bst
    import numpy as np
    from numpy.testing import assert_allclose
    # vle
    with bst.System(algorithm='phenomena based') as sys:
        bst.settings.set_thermo(['Water', 'Ethanol'], cache=True)
        liquid = bst.Stream(Water=50, Ethanol=10, units='kg/hr')
        vapor = bst.Stream(Ethanol=10, Water=50, phase='g', units='kg/hr')
        stage = bst.StageEquilibrium(phases=('g', 'l'), ins=[liquid, vapor])
    sys.simulate()
    streams = stage.ins + stage.outs
    actuals = [i.mol.copy() for i in streams]
    sys.run_phenomena()
    values = [i.mol for i in streams]
    for actual, value in zip(actuals, values):
        assert_allclose(actual, value)
 
def test_trivial_liquid_extraction_case():
    # liquid extraction
    with bst.System(algorithm='phenomena based') as sys:
        bst.settings.set_thermo(['Water', 'Methanol', 'Octanol'], cache=True)
        feed = bst.Stream(Water=500, Methanol=50)
        solvent = bst.Stream(Octanol=500, T=330)
        MSE = bst.MultiStageEquilibrium(N_stages=2, ins=[feed, solvent], phases=('L', 'l'))
    sys.simulate()
    extract, raffinate = MSE.outs
    actual = extract.imol['Methanol'] / feed.imol['Methanol']
    T_actual = extract.T
    sys.run_phenomena()
    value = extract.imol['Methanol'] / feed.imol['Methanol']
    T = extract.T
    assert_allclose(actual, value, rtol=1e-3, atol=1e-3)
    assert_allclose(T_actual, T, rtol=1e-3, atol=1e-3)

def test_trivial_distillation_case():   
    import biosteam as bst
    import thermosteam as tmo
    import numpy as np
    from numpy.testing import assert_allclose
    # distillation
    with bst.System(algorithm='phenomena based') as sys:
        bst.settings.set_thermo(['Water', 'Ethanol'], cache=True)
        feed = bst.Stream(Ethanol=80, Water=100, T=353.455)
        MSE = bst.MultiStageEquilibrium(N_stages=5, ins=[feed], feed_stages=[2],
            outs=['vapor', 'liquid'],
            stage_specifications={0: ('Reflux', 0.673), -1: ('Boilup', 2.57)},
            phases=('g', 'l'),
            use_cache=True,
            maxiter=200,
        )
        MSE.molar_tolerance = 1e-9
        MSE.relative_molar_tolerance = 1e-9
    sys.simulate()
    vapor, liquid = MSE.outs
    actual = round(vapor.imol['Ethanol'] / feed.imol['Ethanol'], 2)
    sys.run_phenomena()
    assert round(vapor.imol['Ethanol'] / feed.imol['Ethanol'], 2) == actual
    
def test_simple_acetic_acid_separation_no_recycle():
    bst.settings.set_thermo(['Water', 'AceticAcid', 'EthylAcetate'], cache=True)
    @bst.SystemFactory
    def system(ins, outs):
        feed = bst.Stream(AceticAcid=6660, Water=43600)
        solvent = bst.Stream(EthylAcetate=65000)
        LE = bst.MultiStageEquilibrium(
            N_stages=6, ins=[feed, solvent], phases=('L', 'l'),
            maxiter=200,
            use_cache=True,
        )
        # DAA = bst.MultiStageEquilibrium(N_stages=6, ins=[LE-0], feed_stages=[3],
        #     outs=['vapor', 'liquid'],
        #     stage_specifications={0: ('Reflux', 0.673), -1: ('Boilup', 2.57)},
        #     phases=('g', 'l'),
        #     method='anderson',
        #     use_cache=True,
        # )
        DEA = bst.MultiStageEquilibrium(N_stages=6, ins=[LE-1], feed_stages=[3],
            outs=['vapor', 'liquid'],
            stage_specifications={0: ('Reflux', 0.673), -1: ('Boilup', 2.57)},
            phases=('g', 'l'),
            use_cache=True,
        )
    init_sys = system()
    init_sys.simulate()
    po = system(algorithm='phenomena based', 
                    molar_tolerance=1e-9,
                    relative_molar_tolerance=1e-9,
                    method='fixed-point')
    sm = system(algorithm='sequential modular',
                    molar_tolerance=1e-9,
                    relative_molar_tolerance=1e-9,
                    method='fixed-point')
    
    for i in range(2): 
        po.simulate()
        po.run_phenomena()
    for i in range(2): 
        sm.simulate()
    
    for s_sm, s_dp in zip(sm.streams, po.streams):
        actual = s_sm.mol
        value = s_dp.mol
        assert_allclose(actual, value, rtol=1e-3, atol=1e-3)

def test_simple_acetic_acid_separation_with_recycle():
    import biosteam as bst
    import numpy as np
    from numpy.testing import assert_allclose
    bst.settings.set_thermo(['Water', 'AceticAcid', 'EthylAcetate'], cache=True)
    solvent_feed_ratio = 1
    @bst.SystemFactory
    def system(ins, outs):
        feed = bst.Stream('feed', AceticAcid=6660, Water=43600)
        solvent = bst.Stream('solvent', EthylAcetate=65000)
        recycle = bst.Stream('recycle')
        LE = bst.MultiStageEquilibrium(
            N_stages=6, ins=[feed, solvent, recycle],
            feed_stages=(0, -1, -1),
            phases=('L', 'l'),
            maxiter=200,
            use_cache=True,
        )
        # DAA = bst.MultiStageEquilibrium(N_stages=6, ins=[LE-0], feed_stages=[3],
        #     outs=['vapor', 'liquid'],
        #     stage_specifications={0: ('Reflux', 0.673), -1: ('Boilup', 2.57)},
        #     maxiter=200,
        #     phases=('g', 'l'),
        #     use_cache=True,
        # )
        DEA = bst.MultiStageEquilibrium(N_stages=6, ins=[LE-1], feed_stages=[3],
            outs=['vapor', 'liquid'],
            stage_specifications={0: ('Reflux', 0.673), -1: ('Boilup', 2.57)},
            phases=('g', 'l'),
            maxiter=200,
            use_cache=True,
        )
        HX = bst.SinglePhaseStage(ins=DEA-0, outs=recycle, T=320, phase='l')
        
        chemicals = bst.settings.chemicals
        
        @LE.add_specification(run=True)
        def fresh_solvent_flow_rate():
            broth = feed.F_mass
            EtAc_recycle = recycle.imass['EthylAcetate']
            EtAc_required = broth * solvent_feed_ratio
            if EtAc_required < EtAc_recycle:
                recycle.F_mass *= EtAc_required / EtAc_recycle
                EtAc_recycle = recycle.imass['EthylAcetate']
            EtAc_fresh = EtAc_required - EtAc_recycle
            solvent.imass['EthylAcetate'] = max(
                0, EtAc_fresh
            )
        
        @solvent.material_balance
        def fresh_solvent_flow_rate():
            s = np.ones(chemicals.size)
            r = np.zeros(chemicals.size)
            v = r.copy()
            index = chemicals.index('EthylAcetate')
            r[index] = 1
            v[index] = solvent_feed_ratio * feed.F_mass / chemicals.EthylAcetate.MW
            return (
                {solvent: s,
                 recycle: r},
                 v
            )
        
    init_sys = system(method='fixed-point')
    init_sys.simulate()
    po = system(algorithm='phenomena based', 
                    molar_tolerance=1e-9,
                    relative_molar_tolerance=1e-9,
                    maxiter=2000,
                    method='fixed-point')
    sm = system(algorithm='sequential modular',
                    molar_tolerance=1e-9,
                    relative_molar_tolerance=1e-9,
                    method='fixed-point')
    time = bst.Timer()
    
    time.start()
    for i in range(1): po.simulate()
    t_phenomena = time.elapsed_time
    
    time.start()
    for i in range(1): sm.simulate()
    t_sequential = time.elapsed_time
    
    for s_sm, s_dp in zip(sm.streams, po.streams):
        actual = s_sm.mol
        value = s_dp.mol
        assert_allclose(actual, value, rtol=1e-6, atol=1e-6)

def test_vlle_case():
    import numpy as np
    import biosteam as bst
    import thermosteam as tmo
    from numpy.testing import assert_allclose
    tmo.settings.set_thermo(['Water', 'Ethanol', 'Octane'], cache=True)
    chemicals = tmo.settings.chemicals
    @bst.SystemFactory
    def system(ins, outs):
        T = 351
        P = 101325
        extract = tmo.Stream('extract')
        raffinate = tmo.Stream('raffinate')
        vapor = tmo.Stream('vapor')
        feed = tmo.Stream('feed', Water=1, Ethanol=0.5, Octane=2, T=T, P=P)
        lle_stage = bst.StageEquilibrium(
            phases=('L', 'l'), 
            ins=[extract, raffinate, feed],
            T=T, P=P,
        )
        vle_L_stage = bst.StageEquilibrium(
            phases=('g', 'l'), 
            ins=[vapor, lle_stage.extract],
            outs=['', extract],
            T=T, P=P,
        )
        vle_l_stage = bst.StageEquilibrium(
            phases=('g', 'l'), 
            ins=[vle_L_stage.vapor, lle_stage.raffinate],
            outs=[vapor, raffinate],
            T=T, P=P,
        )
        total_components = feed.mol.to_array()
        @lle_stage.add_specification(run=True)
        def spec():
            recycles = (extract.mol, raffinate.mol, vapor.mol)
            total_recycles = sum(recycles)
            if total_recycles.all():
                feed.mol[:] = 0
                factor = total_components / total_recycles
                for i in recycles: i *= factor
            else:
                mol = (
                    total_components 
                    - extract.mol
                    - raffinate.mol
                    - vapor.mol
                )
                mol[mol < 0] = 0
                feed.mol = mol
        
        original = lle_stage._create_material_balance_equations
        def spec(composition_sensitive):
            feed.mol[:] = 0
            N = chemicals.size
            e = np.ones(N)
            r = np.ones(N)
            v = np.ones(N)
            new_material_balance = (
                {extract: e,
                 raffinate: r,
                 vapor: v},
                 total_components
            )
            if composition_sensitive:
                return original(composition_sensitive)
            else:
                balances = original(composition_sensitive)
                balances[0] = new_material_balance
                return balances
            
        lle_stage._create_material_balance_equations = spec
    
    init_sys = system(method='fixed-point')
    init_sys.simulate()
    po = system(algorithm='phenomena based', 
                    molar_tolerance=1e-9,
                    relative_molar_tolerance=1e-9,
                    maxiter=2000,
                    method='fixed-point')
    sm = system(algorithm='sequential modular',
                    molar_tolerance=1e-9,
                    relative_molar_tolerance=1e-9,
                    method='fixed-point')
    time = bst.Timer()
    
    time.start()
    for i in range(1): po.simulate()
    t_phenomena = time.elapsed_time
    
    time.start()
    for i in range(1): sm.simulate()
    t_sequential = time.elapsed_time
    
    for s_sm, s_dp in zip(sm.streams, po.streams):
        actual = s_sm.mol
        value = s_dp.mol
        assert_allclose(actual, value, rtol=1e-4, atol=1e-4)
    
    
    # assert_allclose(s.mol, [1, 0.5, 2]) # mass balance
    # total = s.F_mol
    # xl = s.imol['l'].sum() / total
    # xL = s.imol['L'].sum() / total if 'L' in s.phases else 0.
    # xg = s.imol['g'].sum() / total
    # assert_allclose(xl, 0.2748928033836762, atol=2e-3, rtol=2e-3)
    # assert_allclose(xL, 0.6370268977038833, atol=2e-3, rtol=2e-3) # mass balance
    # assert_allclose(xg, 0.08808029891244049, atol=2e-3, rtol=2e-3) # mass balance

if __name__ == '__main__':
    # test_trivial_lle_case()
    # test_trivial_vle_case()
    test_trivial_liquid_extraction_case()
    test_trivial_distillation_case()
    test_simple_acetic_acid_separation_no_recycle()
    test_simple_acetic_acid_separation_with_recycle()
    test_vlle_case()


# -*- coding: utf-8 -*-
"""
General functional algorithms based on MESH equations to solve multi-stage.
"""
import numpy as np
from numba import njit
from math import inf
from scipy.optimize import minimize, lsq_linear

# %% Diagonal matrix solution

@njit(cache=True)
def solve_tridiagonal_matrix(a, b, c, d): # Tridiagonal matrix solver
    """
    Solve a tridiagonal matrix using Thomas' algorithm.
    
    http://en.wikipedia.org/wiki/Tridiagonal_matrix_algorithm
    http://www.cfd-online.com/Wiki/Tridiagonal_matrix_algorithm_-_TDMA_(Thomas_algorithm)
    
    Notes
    -----
    `a` array starts from a1 (not a0).
    
    """
    n = d.shape[0] - 1 # number of equations minus 1
    for i in range(n):
        inext = i + 1
        m = a[i] / b[i]
        b[inext] -= m * c[i] 
        d[inext] -= m * d[i]
    
    b[n] = d[n] / b[n]
    for i in range(n-1, -1, -1):
        b[i] = (d[i] - c[i] * b[i+1]) / b[i]
    return b

@njit(cache=True)
def solve_block_tridiagonal_matrix(A, B, C, d):
    """
    Solve a block tridiagonal matrix using Thomas' algorithm.
    
    http://en.wikipedia.org/wiki/Tridiagonal_matrix_algorithm
    http://www.cfd-online.com/Wiki/Tridiagonal_matrix_algorithm_-_TDMA_(Thomas_algorithm)
    
    Notes
    -----
    `a` array starts from a1 (not a0).
    
    """
    d = d.copy()
    n = d.shape[0] - 1 # number of equations minus 1
    for i in range(n):
        inext = i + 1
        M = np.linalg.solve(B[i], A[i])
        B[inext] -= M @ C[i] 
        d[inext] -= M @ d[i]
        
    d[n] = np.linalg.solve(B[n], d[n])
    for i in range(n-1, -1, -1):
        d[i] = np.linalg.solve(B[i], d[i] - C[i] @ d[i+1])
    return d
    
@njit(cache=True)
def solve_left_bidiagonal_matrix(a, b, d): # Left bidiagonal matrix solver
    """
    Solve a left bidiagonal matrix using a reformulation of Thomas' algorithm.
    """
    n = d.shape[0] - 1 # number of equations minus 1
    for i in range(n):
        inext = i + 1
        m = a[i] / b[i]
        d[inext] = d[inext] - m * d[i]
    
    b[n] = d[n] / b[n]

    for i in range(n-1, -1, -1):
        b[i] = d[i] / b[i]
    return b

@njit(cache=True)
def solve_right_bidiagonal_matrix(b, c, d): # Right bidiagonal matrix solver
    """
    Solve a right bidiagonal matrix using a reformulation of Thomas' algorithm.
    """
    n = d.shape[0] - 1 # number of equations minus 1
    b[n] = d[n] / b[n]

    for i in range(n-1, -1, -1):
        b[i] = (d[i] - c[i] * b[i+1]) / b[i]
    return b

# %% Bulk flow solution

@njit(cache=True)
def hot_start_top_flow_rates(
        bottom_flows, phase_ratios, stage_index, top_feed_flows,
        bottom_feed_flows, neg_asplit, neg_bsplit, N_stages,
    ):
    """
    Solve a-phase flow rates for a single component across 
    equilibrium stages with side draws. 

    Parameters
    ----------
    bottom_flows : Iterable[1d array]
        Bottom flow rates by stages.
    phase_ratios : 1d array
        Phase ratios by stage. The phase ratio for a given stage is 
        defined as F_a / F_b; where F_a and F_b are the flow rates 
        of phase a (extract or vapor) and b (raffinate or liquid) leaving the stage 
        respectively.
    stage_index : 1d array
        Stage index for phase ratios.
    top_feed_flows : Iterable[1d array]
        Top flow rates of all components fed across stages. Shape should be 
        (N_stages, N_chemicals).
    bottom_feed_flows : Iterable [1d array]
        Bottom flow rates of all components fed across stages. Shape should be 
        (N_stages, N_chemicals).
    neg_asplit : 1d array
        Negative of the fraction of phase a being sent to the next stage.
    neg_bsplit : 1d array
        Negative of the fraction of phase b being sent to the next stage.

    Returns
    -------
    flow_rates_a: 2d array
        Flow rates of phase a with stages by row and components by column.

    """
    d = top_feed_flows.copy()
    b = d.copy()
    c = d.copy()
    for i in range(N_stages): 
        c[i] = neg_asplit[i]
        b[i] = 1
    for n in range(stage_index.size):
        i = stage_index[n]
        B = phase_ratios[n]
        if B <= 1e-32:
            b[i] = inf
        else:
            b[i] += 1 / B 
        if i == 0:
            d[i] += bottom_feed_flows[i]
        else:
            d[i] += bottom_feed_flows[i] - bottom_flows[i - 1] * neg_bsplit[i - 1]
    return solve_right_bidiagonal_matrix(b, c, d)

@njit(cache=True)
def hot_start_bottom_flow_rates(
        top_flows, phase_ratios, stage_index, top_feed_flows,
        bottom_feed_flows, neg_asplit, neg_bsplit, N_stages
    ):
    """
    Solve a-phase flow rates for a single component across 
    equilibrium stages with side draws. 

    Parameters
    ----------
    bottom_flows : Iterable[1d array]
        Bottom flow rates by stages.
    phase_ratios : 1d array
        Phase ratios by stage. The phase ratio for a given stage is 
        defined as F_a / F_b; where F_a and F_b are the flow rates 
        of phase a (extract or vapor) and b (raffinate or liquid) leaving the stage 
        respectively.
    stage_index : 1d array
        Stage index for phase ratios.
    top_feed_flows : Iterable[1d array]
        Top flow rates of all components fed across stages. Shape should be 
        (N_stages, N_chemicals).
    bottom_feed_flows : Iterable [1d array]
        Bottom flow rates of all components fed across stages. Shape should be 
        (N_stages, N_chemicals).
    neg_asplit : 1d array
        Negative of the fraction of phase a being sent to the next stage.
    neg_bsplit : 1d array
        Negative of the fraction of phase b being sent to the next stage.

    Returns
    -------
    flow_rates_a: 2d array
        Flow rates of phase a with stages by row and components by column.

    """
    d = bottom_feed_flows.copy()
    b = d.copy()
    a = d.copy()
    for i in range(N_stages): 
        a[i] = neg_bsplit[i]
        b[i] = 1
    last_stage = N_stages - 1
    for n in range(stage_index.size):
        i = stage_index[n]
        b[i] += phase_ratios[n]
        if i == last_stage:
            d[i] += top_feed_flows[i]
        else:
            d[i] += top_feed_flows[i] - top_flows[i + 1] * neg_asplit[i + 1]
    return solve_left_bidiagonal_matrix(a, b, d)

@njit(cache=True)
def create_block_tridiagonal_matrix(As, Bs, Cs):
    """
    Create a dense block tridiagonal matrix.

    Parameters
    ----------
    As : list of (m x m) arrays
        Lower-diagonal blocks, length n-1
    Bs : list of (m x m) arrays
        Diagonal blocks, length n
    Cs : list of (m x m) arrays
        Upper-diagonal blocks, length n-1

    Returns
    -------
    M : (n*m, n*m) ndarray
        Dense block tridiagonal matrix
    """
    n = len(Bs)
    m = Bs[0].shape[0]
    n_vars = n * m
    M = np.zeros((n_vars, n_vars))

    # initialize first row slice
    row_start = 0
    row_end = row_start + m
    row_end_next = row_end + m
    row_slice = slice(row_start, row_end)
    last = n-1
    for i in range(n):
        M[row_slice, row_slice] = Bs[i]  # diagonal
        if i == last: break
        next_slice = slice(row_end, row_end_next)
        M[row_slice, next_slice] = Cs[i]  # upper
        M[next_slice, row_slice] = As[i]  # lower
        # prepare for next iteration
        row_slice = next_slice
        row_start = row_end
        row_end = row_end_next
        row_end_next += m
    return M

# %% Component flow rate solution

@njit(cache=True)
def bottom_flow_rates(
        stripping_factors, 
        feed_flows,
        neg_asplit,
        neg_bsplit,
        N_stages,
    ):
    """
    Solve a-phase flow rates for a single component across equilibrium stages with side draws. 

    Parameters
    ----------
    stripping_factors : Iterable[1d array]
        The ratio of component flow rates in phase a (extract or vapor) over
        the component flow rates in phase a (raffinate or liquid). 
    feed_flows : Iterable[1d array]
        Flow rates of all components fed across stages. Shape should be 
        (N_stages, N_chemicals).
    neg_asplit : 1d array
        Negative of the fraction of phase a being sent to the next stage.
    neg_bsplit : 1d array
        Negative of the fraction of phase b being sent to the next stage.

    Returns
    -------
    flow_rates_a : 2d array
        Flow rates of phase a with stages by row and components by column.

    """
    b = 1. + stripping_factors
    c = np.expand_dims(neg_asplit[1:], -1) * stripping_factors[1:]
    d = feed_flows.copy()
    a = neg_bsplit
    return solve_tridiagonal_matrix(a, b, c, d) 

@njit(cache=True)
def liquid_compositions(
        bulk_vapor_flow_rates,
        bulk_liquid_flow_rates,
        partition_coefficients, 
        feed_flows,
        neg_asplit,
        neg_bsplit,
        N_stages,
    ):
    """
    Solve a-phase flow rates for a single component across equilibrium stages with side draws. 

    Parameters
    ----------
    stripping_factors : Iterable[1d array]
        The ratio of component flow rates in phase a (extract or vapor) over
        the component flow rates in phase a (raffinate or liquid). 
    feed_flows : Iterable[1d array]
        Flow rates of all components fed across stages. Shape should be 
        (N_stages, N_chemicals).
    neg_asplit : 1d array
        Negative of the fraction of phase a being sent to the next stage.
    neg_bsplit : 1d array
        Negative of the fraction of phase b being sent to the next stage.

    Returns
    -------
    flow_rates_a : 2d array
        Flow rates of phase a with stages by row and components by column.

    """
    KV = bulk_vapor_flow_rates * partition_coefficients
    b = bulk_liquid_flow_rates + KV
    c = np.expand_dims(neg_asplit[1:], -1) * KV[1:]
    d = feed_flows.copy()
    a = np.expand_dims(neg_bsplit, -1) * bulk_liquid_flow_rates
    return solve_tridiagonal_matrix(a, b, c, d)

@njit(cache=True)
def bottom_flows_mass_balance(
        top_flows, feed_flows, asplit, bsplit,
        N_stages
    ):
    bottom_flows = 0 * top_flows
    bottom_flows[0] = feed_flows[0] + top_flows[1] * asplit[1] - top_flows[0]
    for i in range(1, N_stages-1):
        bottom_flows[i] = (
            feed_flows[i] + bsplit[i-1] * bottom_flows[i-1] + 
            top_flows[i+1] * asplit[i+1] - top_flows[i]
        )
    bottom_flows[-1] = feed_flows[-1] + bsplit[-2] * bottom_flows[-2] - top_flows[-1]
    return bottom_flows


@njit(cache=True)
def top_flows_mass_balance(
        bottom_flows, feed_flows, asplit, bsplit, N_stages
    ):
    top_flows = 0 * bottom_flows
    top_flows[-1] = feed_flows[-1] + bsplit[-2] * bottom_flows[-2] - bottom_flows[-1]
    for i in range(N_stages-2, 0, -1):
        top_flows[i] = (
            feed_flows[i] + bsplit[i-1] * bottom_flows[i-1] + 
            top_flows[i+1] * asplit[i+1] - bottom_flows[i]
        )
    top_flows[0] = feed_flows[0] + top_flows[1] * asplit[1] - bottom_flows[0]
    return top_flows

# %% Energy balance solution

@njit(cache=True)
def bulk_vapor_and_liquid_flow_rates(
        hl, hv, 
        neg_asplit, neg_bsplit, 
        top_split, bottom_split, 
        N_stages, H_feeds, F_feeds,
        variables,
        values,
        bulk_feed,
    ):
    # Equations:
    # - hl0*L0 + hv1*V1 + hl1*L1 - hv2*V2 = Q1
    # - L0 + V1 + L1 - V2 = 0
    N_bulk = 2 * N_stages
    Q = np.zeros(N_bulk)
    L0 = -1 # L0
    V1 = 0 # V1
    L1 = 1 # L1
    V2 = 2 # V2
    A = np.zeros((N_bulk, N_bulk))
    last = N_stages - 1
    for i in range(N_stages):
        variable = variables[i]
        value = values[i] 
        n = 2 * i
        m = n + 1
        ilast = i - 1
        inext = i + 1
        
        # Bulk material balance
        if L0 != -1: A[m, L0] = neg_bsplit[ilast]
        A[m, V1] = 1
        A[m, L1] = 1
        if i != last: A[m, V2] = neg_asplit[inext]
        Q[m] = F_feeds[i]
        
        # Energy balance
        if variable == 'Q':
            if L0 != -1: A[n, L0] = hl[ilast] * neg_bsplit[ilast]
            A[n, V1] = hv[i]
            A[n, L1] = hl[i]
            Q[n] = H_feeds[i] + value
            if i != last: A[n, V2] = hv[inext] * neg_asplit[inext]
        elif variable == 'B':
            A[n, V1] = -1
            A[n, L1] = value
        elif variable == 'F':
            A[n, L1] = 1
            Q[n] = bulk_feed * value
        else:
            raise ValueError('unexpected variable specification')
        L0 += 2
        V1 += 2
        L1 += 2
        V2 += 2
    VLs = np.linalg.solve(A, Q)
    VLs[VLs < 1e-16] = 1e-16
    return VLs[::2], VLs[1::2] # V, L

@njit(cache=True)
def phase_ratio_departures(
        L, V, hl, hv, neg_asplit, asplit, bsplit, 
        N_stages,
        specification_index,
        H_feeds,
        RF=False, # Reflux rate and bottoms flow rate specification
    ):
    # hV1*L1*dB1 - hv2*L2*dB2 = Q1 + H_in - H_out
    b = hv * L
    c = b[1:] * neg_asplit[1:]
    Hl_out = hl * L
    Hv_out = hv * V
    d = H_feeds - Hl_out - Hv_out
    Hl_in = (Hl_out * bsplit)[:-1]
    Hv_in = (Hv_out * asplit)[1:]
    d[1:] += Hl_in
    d[:-1] += Hv_in
    if RF:
        dB = d * 0
        dB[2:] = solve_left_bidiagonal_matrix(b[2:], c[1:], d[1:-1])
        return dB
    else:
        n_last = d.size - 1
        for j in specification_index:
            b[j] = 1
            d[j] = 0
            jlast = j - 1
            if jlast > 0: c[jlast] = 0
            if j < n_last: c[j] = 0
        return solve_right_bidiagonal_matrix(b, c, d)

@njit(cache=True)
def temperature_departures(Cv, Cl, Hv, Hl, asplit, bsplit,
                           N_stages, H_feeds):
    # ENERGY BALANCE
    # C1dT1 - Cv2*dT2 - Cl0*dT0 = Q1 - H_out + H_in
    b = (Cv + Cl)
    a = -(Cl * bsplit)
    c = -(Cv * asplit)[1:]
    d = H_feeds - Hl - Hv
    d[1:] += (Hl * bsplit)[:-1]
    d[:-1] += (Hv * asplit)[1:]
    return solve_tridiagonal_matrix(a, b, c, d)

def get_neighbors(index=None, all_index=None, missing=None, size=None):
    if size is not None:
        all_index = set(range(size))
    elif all_index is not None:
        all_index = set(all_index)
    if sum([i is None for i in (index, all_index, missing)]) > 1:
        raise ValueError('at least two arguments must be given')
    if missing is None: 
        missing = all_index.difference(index)
    else:
        missing = set(missing)
    if index is None:
        index_set = all_index.difference(missing)
    else:
        index_set = set(index)
    if all_index is None:
        all_index = index | missing
    size = len(all_index)
    neighbors = []
    for i in missing:
        lb = i
        while lb > -1:
            lb -= 1
            if lb in index_set: break
        ub = i
        while ub < size:
            ub += 1
            if ub in index_set: break
        if ub == size:
            neighbors.append(
                (i, (lb,))
            )
        elif lb == -1:
            neighbors.append(
                (i, (ub,))
            )
        else:
            neighbors.append(
                (i, (lb, ub))
            )
    return neighbors

def expand(values, index, size):
    new_values = np.zeros(size)
    new_values[index] = values
    return new_values

def fillmissing(all_neighbors, values):
    for i, neighbors in all_neighbors:
        if len(neighbors) == 2:
            lb, ub = neighbors
            lb_distance = i - lb
            ub_distance = ub - i
            sum_distance = lb_distance + ub_distance
            wlb = ub_distance / sum_distance
            wub = lb_distance / sum_distance
            x = wlb * values[lb] + wub * values[ub]
            values[i] = x
        else:
            values[i] = values[neighbors[0]]
    return values

    
    


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from . import agitator
from . import vacuum
from . import pressure_vessel
from . import flash_vessel_design
from . import cost_index
from . import batch
from . import specification_factors
from . import column_design
from . import heat_transfer
from . import tank_design
from . import geometry
from . import mechanical
from . import utils
from . import MESH

__all__ = (
    'MESH',
    *agitator.__all__,
    *cost_index.__all__,
    *pressure_vessel.__all__,
    *vacuum.__all__,
    *flash_vessel_design.__all__,
    *batch.__all__,
    *specification_factors.__all__,
    *column_design.__all__,
    *heat_transfer.__all__,
    *tank_design.__all__,
    *mechanical.__all__,
    *utils.__all__,
    *geometry.__all__,
)

from .agitator import *
from .pressure_vessel import *
from .specification_factors import *
from .column_design import *
from .vacuum import *
from .flash_vessel_design import *
from .cost_index import *
from .batch import *
from .heat_transfer import *
from .tank_design import *
from .mechanical import *
from .utils import *
from .geometry import *


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-, Yoel Cortes-Pena <yoelcortes@gmail.com>,
#                      Yalin Li <mailto.yalin.li@gmail.com>,
#                      Sarang Bhagwat <sarangb2@gmail.com>
#
# This module implements a filtering feature from the stats module of the QSDsan library:
# QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems
# Copyright (C) 2020-, Yalin Li <mailto.yalin.li@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.

from scipy.spatial.distance import cdist
from scipy.optimize import shgo, differential_evolution
import numpy as np
import pandas as pd
from chaospy import distributions as shape
from ._indicator import Indicator
from ._feature import MockFeature
from ._utils import var_indices, var_columns, indices_to_multiindex
from ._prediction import ConvergenceModel
from .._unit import Unit
from biosteam.exceptions import FailedEvaluation
from warnings import warn
from collections.abc import Sized
from biosteam.utils import Timer
from typing import Optional, Callable
from ._parameter import Parameter
from .evaluation_tools import load_default_parameters
import pickle

__all__ = ('Model', 'EasyInputModel')

def replace_nones(values, replacement):
    for i, j in enumerate(values):
        if j is None: values[i] = replacement
    return values

def codify(statement):
    statement = replace_apostrophes(statement)
    statement = replace_newline(statement)
    return statement

def replace_newline(statement):
    statement = statement.replace('\n', ';')
    return statement

def replace_apostrophes(statement):
    statement = statement.replace('’', "'").replace('‘', "'").replace('“', '"').replace('”', '"')
    return statement

def create_function(code, namespace):
    def wrapper_fn(statement):
        def f(x):
            namespace['x'] = x
            exec(codify(statement), namespace)
        return f
    function = wrapper_fn(code)
    return function

# %% Fix compatibility with new chaospy version

import chaospy as cp
version_components = cp.__version__.split('.')
CP_MAJOR, CP_MINOR = int(version_components[0]), int(version_components[1])
CP4 = (CP_MAJOR, CP_MINOR) >= (4, 0)
if CP4:
    from inspect import signature
    def save_repr_init(f):
        defaults = list(signature(f).parameters.values())[1:]
        defaults = {i.name: i.default for i in defaults}
        def init(self, *args, **kwargs):
            if not hasattr(self, '_repr'):
                self._repr = params = defaults.copy()
                for i, j in zip(params, args): params[i] = j
                params.update(kwargs)
            f(self, *args, **kwargs)
        return init
    
    shapes = cp.distributions
    Distribution = cp.distributions.Distribution
    baseshapes = set([i for i in cp.distributions.baseclass.__dict__.values()
                      if isinstance(i, type) and issubclass(i, Distribution)])
    for i in shapes.__dict__.values():
        if isinstance(i, type) and issubclass(i, Distribution) and i not in baseshapes:
            i.__init__ = save_repr_init(i.__init__)
    del signature, save_repr_init, shapes, baseshapes, Distribution, i
del version_components, CP_MAJOR, CP_MINOR, CP4

# %% Simulation of process systems

class Model:
    """
    Create a Model object that allows for evaluation over a sample space.
    
    Parameters
    ----------
    system : System
        Should reflect the model state.
    indicators : tuple[Indicator]
        Indicators to be evaluated by model.
    specification=None : Function, optional
        Loads specifications once all parameters are set. Specification should 
        simulate the system as well.
    parameters=None : Iterable[Parameter], optional
        Parameters to sample from.
    exception_hook : callable(exception, sample)
        Function called after a failed evaluation. The exception hook should 
        return either None or indicator values given the exception and sample.

    """
    __slots__ = (
        '_system', # [System]
        '_parameters', # list[Parameter] All parameters.
        '_optimized_parameters', # list[Parameter] Parameters being optimized.
        '_specification', # [function] Loads specifications once all parameters are set.
        'table',            # [DataFrame] All arguments and results.
        'retry_evaluation', # [bool] Whether to retry evaluation if it fails
        'convergence_model',# [ConvergenceModel] Prediction model for recycle convergence.
        '_indicators',         # tuple[Indicator] Indicators to be evaluated by model.
        '_index',           # list[int] Order of sample evaluation for performance.
        '_samples',         # [array] Argument sample space.
        '_exception_hook',  # [callable(exception, sample)] Should return either None or indicator value given an exception and the sample.
    )
    default_optimizer_options = {
        'shgo': dict(f_tol=1e-3, minimizer_kwargs=dict(f_tol=1e-3)),
        'differential evolution': {'seed': 0, 'popsize': 12, 'tol': 1e-3}
    }
    default_optimizer = 'shgo'
    default_convergence_model = None # Optional[str] Default convergence model
    load_default_parameters = load_default_parameters
    
    @property
    def specification(self) -> Callable:
        """Process specification."""
        return self._specification
    @specification.setter
    def specification(self, specification):
        if specification:
            if callable(specification):
                self._specification = specification
            else:
                raise AttributeError(
                    "specification must be callable or None; "
                   f"not a '{type(specification).__name__}'"
                )
        else:
            self._specification = None
    
    @property
    def system(self):
        return self._system
    
    def __len__(self):
        return len(self._parameters)
    
    def get_baseline_scenario(self, parameters=None, array=False):
        """Return a pandas Series object of parameter baseline values."""
        if parameters is None: parameters = self._parameters
        sample = [] if array else {}
        for p in parameters:
            baseline = p.baseline
            if baseline is None: raise RuntimeError(f'{p} has no baseline value')  
            if array: sample.append(baseline)
            else: sample[p.index] = baseline
        return np.array(sample) if array else pd.Series(sample)
    get_baseline_sample = get_baseline_scenario
    
    @property
    def optimized_parameters(self):
        """tuple[Parameter] All parameters optimized in the model."""
        return tuple(self._optimized_parameters)
    @optimized_parameters.setter
    def optimized_parameters(self, parameters):
        self._optimized_parameters = parameters = list(parameters)
        isa = isinstance
        for i in parameters:
            assert isa(i, Parameter), 'all elements must be Parameter objects'
        Parameter.check_indices_unique(self.features)
    
    @property
    def parameters(self):
        """tuple[Parameter] All parameters added to the model."""
        return self.get_parameters()
    @parameters.setter
    def parameters(self, parameters):
        self.set_parameters(parameters)
    
    def set_parameters(self, parameters):
        """Set parameters."""
        self._parameters = parameters = list(parameters)
        isa = isinstance
        for i in parameters:
            assert isa(i, Parameter), 'all elements must be Parameter objects'
        Parameter.check_indices_unique(self.features)
    
    def parameters_from_df(self, df_or_filename, namespace=None):
        """
        Load a list (from a DataFrame or spreadsheet) of distributions and statements
        to load values for user-selected parameters.
        
        Parameters
        ----------
        df_or_filename : pandas.DataFrame or file path to a spreadsheet with the following column titles.
                         
            * 'Parameter name' [String] Name of the parameter.
            
            * 'Element' [String] 
                        
            * 'Units' [String] 
                        
            * 'Baseline' [float] The baseline value of the parameter.
                        
            * 'Shape' {'Uniform', 'Triangular'} The shape of the parameter distribution.
                        
            * 'Lower' [float] The lower value defining the shape of the parameter distribution.
                        
            * 'Midpoint' [float] The midpoint value defining the shape of a 'Triangular' parameter distribution.
                        
            * 'Upper' [float] The upper value defining the shape of the parameter distribution.
                        
            * 'Load statement' [String] A statement executed to load the value of the parameter. 
            
        namespace : dict, optional
            Dictionary used to update the namespace accessed when executing
            statements to load values into model parameters. Defaults to the
            system's flowsheet dict.
                        
        """
        
        df = df_or_filename
        if type(df) is not pd.DataFrame:
            try: 
                df = pd.read_excel(df_or_filename)
            except:
                df = pd.read_csv(df_or_filename)
                
        if namespace is None: namespace = {}
        namespace = self.system.flowsheet.to_dict() | namespace
        
        param = self.parameter
        
        for i, row in df.iterrows():
            name = row['Parameter name']
            element = row['Element'] # currently only compatible with String elements
            try:
                coupled = row['Coupled']
            except:
                coupled = row['Kind'] == 'coupled'
            units = row['Units']
            baseline = row['Baseline']
            shape_data = row['Shape']
            lower, midpoint, upper = row['Lower'], row['Midpoint'], row['Upper']
            load_statements = row['Load statement']
            
            D = None
            if shape_data.lower() in ['triangular', 'triangle',]:
                D = shape.Triangle(lower, midpoint, upper)
            elif shape_data.lower() in ['uniform',]:
                if not str(midpoint)=='nan':
                    raise ValueError(f"The parameter distribution for {name} ({element}) is 'Uniform' but was associated with a given midpoint value.")
                D = shape.Uniform(lower, upper)
                
            param(name=name, 
                  setter=create_function(load_statements, namespace), 
                  element=element, 
                  coupled=coupled, 
                  units=units,
                  baseline=baseline, 
                  distribution=D)
            
    def get_parameters(self):
        """Return parameters."""
        return tuple(self._parameters)
    
    def get_joint_distribution(self):
        """
        Return a chaospy joint distribution object constituting of all
        parameter objects.
        
        """
        return cp.distributions.J(*[i.distribution for i in self.get_parameters()])
    
    def get_distribution_summary(self, xlfile=None):
        """Return dictionary of shape name-DataFrame pairs."""
        parameters = self.get_parameters()
        if not parameters: return None
        parameters_by_shape = {}
        shape_keys = {}
        for p in parameters:
            distribution = p.distribution
            if not distribution: continue
            shape = type(distribution).__name__
            if shape in parameters_by_shape:
                parameters_by_shape[shape].append(p)
            else:
                parameters_by_shape[shape] = [p]
                shape_keys[shape] = tuple(distribution._repr.keys())
        tables_by_shape = {}
        for shape, parameters in parameters_by_shape.items():
            data = []
            columns = ('Element', 'Name', 'Units', 'Shape', *shape_keys[shape])
            for p in parameters:
                distribution = p.distribution
                element = p.element_name
                name = p.name.replace(element, '')
                units = p.units or ''
                values = distribution._repr.values()
                data.append((element, name, units, shape, *values))
            tables_by_shape[shape] =  pd.DataFrame(data, columns=columns)
        if xlfile:
            with pd.ExcelWriter(xlfile) as writer:
                for shape, df in tables_by_shape.items():
                    df.to_excel(writer, sheet_name=shape)
        return tables_by_shape    
    
    def optimized_parameter(self, *args, **kwargs):
        return self.parameter(*args, **kwargs, optimized=True, coupled=True)
    
    def parameter(self, 
            setter: Optional[Callable]=None,
            element: Optional[Unit]=None, 
            coupled: Optional[bool]=None,
            name: Optional[str]=None, 
            distribution: Optional[str|shape.baseclass.Distribution]=None, 
            units: Optional[str]=None, 
            baseline: Optional[float]=None, 
            bounds: Optional[tuple[float, float]]=None, 
            hook: Optional[Callable]=None, 
            description: Optional[str]=None, 
            optimized=False, 
            kind=None, # For backwards compatibility
            safe=False,
        ):
        """
        Define and register parameter.
        
        Parameters
        ---------*    
        setter : 
            Should set parameter in the element.
        element : 
            Element in the system being altered.
        coupled : 
            Whether parameter is coupled to the system's mass and energy balances.
            This allows a ConvergenceModel to predict it's impact on recycle loops.
            Defaults to False.
        name : 
            Name of parameter. If None, default to argument name of setter.
        distribution : 
            Parameter distribution.
        units : 
            Parameter units of measure
        baseline : 
            Baseline value of parameter.
        bounds : 
            Lower and upper bounds of parameter.
        hook : 
            Should return the new parameter value given the sample.
        
        """
        if kind == 'coupled': coupled = True
        elif coupled is None: coupled = False
        if isinstance(setter, Parameter):
            if element is None: element = setter.element
            if coupled is None: coupled = setter.coupled
            if name is None: name = setter.name
            if distribution is None: distribution = setter.distribution
            if units is None: units = setter.units
            if baseline is None: baseline = setter.baseline
            if bounds is None: bounds = setter.bounds
            if hook is None: hook = setter.hook
            if description is None: description = setter.description
            setter = setter.setter
        elif isinstance(setter, MockFeature):
            if element is None: element = setter.element
            if name is None: name = setter.name
            if units is None: units = setter.units
        elif not setter:
            return lambda setter: self.parameter(setter, element, coupled, name,
                                                 distribution, units, baseline,
                                                 bounds, hook, description, optimized)
        p = Parameter(name, setter, element,
                      self.system, distribution, units, 
                      baseline, bounds, coupled, hook, description)
        if safe:
            Parameter.check_index_unique(p, self.features)
        else:
            key = (p.element_name, p.name)
            dct = {(i.element_name, i.name): i for i in self.features}
            if key in dct:
                old_p = dct[key]
                try:
                    if optimized:
                        self._optimized_parameters.remove(old_p)
                    else:
                        self._parameters.remove(old_p)
                except:
                    raise ValueError(
                             "each feature must have a unique element and name; "
                            f"feature with element {repr(p.element)} "
                            f"and name {repr(p.name)} already present"
                        )
        if optimized:
            self._optimized_parameters.append(p)
        else:
            self._parameters.append(p)
        return p
    
    def problem(self):
        """
        Return a dictionary of parameter metadata (referred to as "problem") 
        to be used for sampling by ``SALib``.
        
        See Also
        --------
        `SALib basics <https://salib.readthedocs.io/en/latest/basics.html#an-example>`_
        
        """
        params = self.get_parameters()
        return {
            'num_vars': len(params),
            'names': [i.name for i in params],
            'bounds': [i.bounds for i in params]
        }

    def sample(self, N, rule, **kwargs): 
        """
        Return N samples from parameter distribution at given rule.
        
        Parameters
        ----------
        N : int
            Number of samples.
        rule : str
            Sampling rule.
        **kwargs :
            Keyword arguments passed to sampler.
        
        Notes
        -----
        This method relies on the ``chaospy`` library for sampling from 
        distributions, and the``SALib`` library for sampling schemes 
        specific to sensitivity analysis.
        
        For sampling from a joint distribution of all parameters, 
        use the following ``rule`` flags:
        
        +-------+-------------------------------------------------+
        | key   | Description                                     |
        +=======+=================================================+
        | ``C`` | Roots of the first order Chebyshev polynomials. |
        +-------+-------------------------------------------------+
        | ``NC``| Chebyshev nodes adjusted to ensure nested.      |
        +-------+-------------------------------------------------+
        | ``K`` | Korobov lattice.                                |
        +-------+-------------------------------------------------+
        | ``R`` | Classical (Pseudo-)Random samples.              |
        +-------+-------------------------------------------------+
        | ``RG``| Regular spaced grid.                            |
        +-------+-------------------------------------------------+
        | ``NG``| Nested regular spaced grid.                     |
        +-------+-------------------------------------------------+
        | ``L`` | Latin hypercube samples.                        |
        +-------+-------------------------------------------------+
        | ``S`` | Sobol low-discrepancy sequence.                 |
        +-------+-------------------------------------------------+
        | ``H`` | Halton low-discrepancy sequence.                |
        +-------+-------------------------------------------------+
        | ``M`` | Hammersley low-discrepancy sequence.            |
        +-------+-------------------------------------------------+
        
        If sampling for sensitivity analysis, use the following ``rule`` flags:
        
        +------------+--------------------------------------------+
        | key        | Description                                |
        +============+============================================+
        | ``MORRIS`` | Samples for Morris One-at-A-Time (OAT)     |
        +------------+--------------------------------------------+
        | ``RBD``    | Chebyshev nodes adjusted to ensure nested. |
        +------------+--------------------------------------------+
        | ``FAST``   | Korobov lattice.                           |
        +------------+--------------------------------------------+
        | ``SOBOL``  | Classical (Pseudo-) Random samples.        |
        +------------+--------------------------------------------+
        
        Note that only distribution bounds (i.e. lower and upper bounds) are 
        taken into account for sensitivity analysis, the type of distribution
        (e.g., triangle vs. uniform) do not affect the sampling. 
        
        """
        rule = rule.upper()
        if rule in ('C', 'NC', 'K', 'R', 'RG', 'NG', 'L', 'S', 'H', 'M'):
            samples = self.get_joint_distribution().sample(N, rule).transpose()
        else:
            if rule == 'MORRIS':
                from SALib.sample import morris as sampler
            elif rule in ('FAST', 'EFAST'):
                from SALib.sample import fast_sampler as sampler
            elif rule == 'RBD':
                from SALib.sample import latin as sampler
            elif rule == 'SOBOL' or rule == 'SALTELLI':
                from SALib.sample import sobol as sampler
            else:
                raise ValueError(f"invalid rule '{rule}'")
            problem = kwargs.pop('problem') if 'problem' in kwargs else self.problem()
            samples = sampler.sample(problem, N=N, **kwargs)
        return samples
    
    def _objective_function(self, sample, loss, parameters, convergence_model=None, **kwargs):
        for f, s in zip(parameters, sample): 
            f.setter(s if f.scale is None else f.scale * s)
        if convergence_model:
            with convergence_model.practice(sample, parameters):
                self._specification() if self._specification else self._system.simulate(**kwargs)
        return loss()
    
    def _update_state(self, sample, convergence_model=None, **kwargs):
        for i, (f, value) in enumerate(zip(self._parameters, sample)): 
            if f.active: 
                f.setter(value)
                f.last_value = value
            else:
                sample[i] = f.last_value
        if convergence_model:
            with convergence_model.practice(sample):
                return self._specification() if self._specification else self._system.simulate(**kwargs)
        else:
            return self._specification() if self._specification else self._system.simulate(**kwargs)
    
    def _evaluate_sample(self, sample, convergence_model=None, **kwargs):
        state_updated = False
        try:
            self._update_state(sample, convergence_model, **kwargs)
            state_updated = True
            return [i() for i in self.indicators]
        except Exception as exception:
            if self.retry_evaluation and not state_updated:
                self._reset_system()
                try:
                    self._update_state(sample, convergence_model, **kwargs)
                    return [i() for i in self.indicators]
                except Exception as new_exception: 
                    exception = new_exception
            if self._exception_hook: 
                values = self._exception_hook(exception, sample)
                self._reset_system()
                if isinstance(values, Sized) and len(values) == len(self.indicators):
                    return values
                elif values is not None:
                    raise RuntimeError('exception hook must return either None or '
                                       'an array of indicator values for the given sample')
            return self._failed_evaluation()
    
    def __init__(self, system, indicators=None, specification=None, 
                 parameters=None, retry_evaluation=None, exception_hook=None):
        self.specification = specification
        if parameters:
            self.set_parameters(parameters)
        else:
            self._parameters = []
        self._optimized_parameters = []
        self._system = system
        self._specification = specification
        self.indicators = indicators or ()
        self.exception_hook = 'warn' if exception_hook is None else exception_hook 
        self.retry_evaluation = bool(system) if retry_evaluation is None else retry_evaluation
        self.table = None
        self._erase()
        
    def copy(self):
        """Return copy."""
        copy = self.__new__(type(self))
        copy._parameters = self._parameters.copy()
        copy._optimized_parameters = self._optimized_parameters.copy()
        copy._system = self._system
        copy._specification = self._specification
        copy._indicators = self._indicators
        if self.table is None:
            copy._samples = copy.table = None
        else:
            copy.table = self.table.copy()
            copy._samples = self._samples
            copy._index = self._index
        return copy
    
    def _erase(self):
        """Erase cached data."""
        self._samples = None
    
    @property
    def exception_hook(self):
        """
        [callable(exception, sample)] Function called after a failed 
        evaluation. The exception hook should return either None or indicator 
        values given the exception and sample.
        
        """
        return self._exception_hook
    @exception_hook.setter
    def exception_hook(self, exception_hook):
        if not exception_hook or callable(exception_hook):
            self._exception_hook = exception_hook
            return
        if isinstance(exception_hook, str):
            if exception_hook == 'ignore':
                self._exception_hook = lambda exception, sample: None
            elif exception_hook == 'warn':
                self._exception_hook = lambda exception, sample: warn(FailedEvaluation(f"[{type(exception).__name__}] {exception}"), stacklevel=6)
            elif exception_hook == 'raise':
                def raise_exception(exception, sample): raise exception from None
                self._exception_hook = raise_exception
            else:
                raise ValueError(f"invalid exception hook name '{exception_hook}'; "
                                  "valid names include 'ignore', 'warn', and 'raise'")
        else:
            raise ValueError('exception hook must be either a callable, a string, or None')
    
    @property
    def indicators(self):
        """tuple[Indicator] Indicators to be evaluated by model."""
        return tuple(self._indicators)
    @indicators.setter
    def indicators(self, indicators):
        self._indicators = indicators = list(indicators)
        isa = isinstance
        for i in indicators:
            if not isa(i, Indicator):
                raise ValueError(f"indicators must be '{Indicator.__name__}' "
                                 f"objects, not '{type(i).__name__}'")
        Indicator.check_indices_unique(self.features)
    
    # Backwards compatibility
    metrics = indicators
    @property
    def _metrics(self): return self._indicators
    @_metrics.setter
    def _metrics(self, metrics): self._indicators = metrics
    
    @property
    def features(self):
        return (*self._parameters, *self._optimized_parameters, *self._indicators)
    
    def indicator(self, getter=None, name=None, units=None, element=None, safe=False):
        """
        Define and register indicator.
        
        Parameters
        ----------    
        getter : function, optional
                 Should return indicator.
        name : str, optional
               Name of indicator. If None, defaults to the name of the getter.
        units : str, optional
                Indicator units of measure
        element : object, optional
                  Element being evaluated. Works mainly for bookkeeping. 
                  Defaults to 'Biorefinery'.
        
        Notes
        -----
        This method works as a decorator.
        
        """
        if isinstance(getter, Indicator):
            if name is None: name = getter.name
            if units is None: units = getter.units
            if element is None: element = getter.element
            getter = getter.getter
        elif isinstance(getter, MockFeature):
            if element is None: element = getter.element
            if name is None: name = getter.name
            if units is None: units = getter.units
        elif not getter: 
            return lambda getter: self.indicator(getter, name, units, element)
        indicator = Indicator(name, getter, units, element)
        if safe:
            Indicator.check_index_unique(indicator, self.features, safe)
        else:
            key = (indicator.element_name, indicator.name)
            dct = {(i.element_name, i.name): i for i in self.features}
            if key in dct:
                old_indicator = dct[key]
                try:
                    self._indicators.remove(old_indicator)
                except:
                    raise ValueError(
                             "each feature must have a unique element and name; "
                            f"feature with element {repr(indicator.element)} "
                            f"and name {repr(indicator.name)} already present"
                        )
        self._indicators.append(indicator)
        return indicator 
    metric = indicator
    
    def _sample_hook(self, samples, parameters):
        if any([p.hook for p in parameters]):
            return np.array(
                [[(i if p.hook is None else p.hook(i))
                  for p, i in zip(parameters, row)]
                 for row in samples]
            )
        else:
            return samples
    
    def _load_sample_order(self, samples, parameters, distance, algorithm=None):
        """
        Sort simulation order to optimize convergence speed
        by minimizing perturbations to the system between simulations.
        
        """
        N_samples = samples.shape[0]
        if N_samples < 2: 
            self._index = list(range(N_samples))
            return
        if distance is None: distance = 'cityblock'
        length = samples.shape[0]
        columns = [i for i, parameter in enumerate(parameters) if parameter.coupled]
        samples = samples.copy()
        samples = samples[:, columns]
        samples_min = samples.min(axis=0)
        samples_max = samples.max(axis=0)
        samples_diff = samples_max - samples_min
        normalized_samples = (samples - samples_min) / samples_diff
        nearest_arr = cdist(normalized_samples, normalized_samples, metric=distance)
        if algorithm is None: algorithm = 'nearest neighbor'
        if algorithm == 'nearest neighbor':
            nearest_arr = np.argsort(nearest_arr, axis=1)
            remaining = set(range(length))
            self._index = index = [0]
            nearest = nearest_arr[0, 1]
            index.append(nearest)
            remaining.remove(0)
            remaining.remove(nearest)
            N_remaining = length - 2
            N_remaining_last = N_remaining + 1
            while N_remaining:
                assert N_remaining_last != N_remaining, "issue in sorting algorithm"
                N_remaining_last = N_remaining
                for i in range(1, length):
                    next_nearest = nearest_arr[nearest, i]
                    if next_nearest in remaining:
                        nearest = next_nearest
                        remaining.remove(nearest)
                        index.append(nearest)
                        N_remaining -= 1
                        break
        else:
            raise ValueError(f"algorithm {algorithm!r} is not available (yet); "
                              "only 'nearest neighbor' is available")
        
    def load_samples(self, samples=None, sort=None, file=None, 
                     autoload=None, autosave=None, distance=None):
        """
        Load samples for evaluation.
        
        Parameters
        ----------
        samples : numpy.ndarray, dim=2, optional
            All parameter samples to evaluate.
        sort : bool, optional
            Whether to internally sort the samples to optimize convergence speed
            by minimizing perturbations to the system between simulations. The
            optimization problem is equivalent to the travelling salesman problem;
            each scenario of (normalized) parameters represent a point in the path.
            Defaults to True.
        file : str, optional
            File to load/save samples and simulation order to/from.
        autosave : bool, optional
            Whether to save samples and simulation order to file (when not loaded from file).
        autoload : bool, optional
            Whether to load samples and simulation order from file (if possible).
        distance : str, optional
            Distance indicator used for sorting. Defaults to 'cityblock'.
            See scipy.spatial.distance.cdist for options.
        algorithm : str, optional
            Algorithm used for sorting. Defaults to 'nearest neighbor'.
            Note that neirest neighbor is a greedy algorithm that is known to result, 
            on average, in paths 25% longer than the shortest path.
        
        """
        parameters = self._parameters
        if autoload:
            try:
                with open(file, "rb") as f: (self._samples, self._index) = pickle.load(f)
            except FileNotFoundError: pass
            else:
                if (samples is None or (samples.shape == self._samples.shape and (samples == self._samples).all())):
                    indicators = self._indicators
                    empty_indicator_data = np.zeros((len(samples), len(indicators)))
                    self.table = pd.DataFrame(np.hstack((samples, empty_indicator_data)),
                                              columns=var_columns(parameters + indicators),
                                              dtype=float)
                    return 
        if not isinstance(samples, np.ndarray):
            raise TypeError(f'samples must be an ndarray, not a {type(samples).__name__} object')
        if samples.ndim == 1:
            samples = samples[:, np.newaxis]
        elif samples.ndim != 2:
            raise ValueError('samples must be 2 dimensional')
        N_parameters = len(parameters)
        if samples.shape[1] != N_parameters:
            raise ValueError(f'number of parameters in samples ({samples.shape[1]}) must be equal to the number of parameters ({N_parameters})')
        indicators = self._indicators
        samples = self._sample_hook(samples, parameters)
        if sort is None: sort = True
        if sort and any([i.coupled for i in parameters]): 
            self._load_sample_order(samples, parameters, distance)
        else:
            self._index = list(range(samples.shape[0]))
        empty_indicator_data = np.zeros((len(samples), len(indicators)))
        self.table = pd.DataFrame(np.hstack((samples, empty_indicator_data)),
                                  columns=var_columns(parameters + indicators),
                                  dtype=float)
        self._samples = samples
        if autosave:
            obj = (samples, self._index)
            try:
                with open(file, 'wb') as f: pickle.dump(obj, f)
            except FileNotFoundError:
                import os
                head, tail = os.path.split(file)
                os.mkdir(head)
                with open(file, 'wb') as f: pickle.dump(obj, f)
            
    def single_point_sensitivity(self, 
            etol=0.01, array=False, parameters=None, indicators=None, evaluate=None, 
            **kwargs
        ):
        if parameters is None: parameters = self.parameters
        bounds = [i.bounds for i in parameters]
        sample = [i.baseline for i in parameters]
        N_parameters = len(parameters)
        index = range(N_parameters)
        if indicators is None: indicators = self.indicators
        N_indicators = len(indicators)
        values_lb = np.zeros([N_parameters, N_indicators])
        values_ub = values_lb.copy()
        if evaluate is None: evaluate = self._evaluate_sample
        baseline_1 = np.array(evaluate(sample, **kwargs))
        sys = self.system
        if not sys.isdynamic: kwargs['recycle_data'] = sys.get_recycle_data()
        for i in index:
            sample_lb = sample.copy()
            sample_ub = sample.copy()
            lb, ub = bounds[i]
            p = parameters[i]
            hook = p.hook
            if hook is not None:
                lb = hook(lb)
                ub = hook(ub)
            sample_lb[i] = lb
            sample_ub[i] = ub
            values_lb[i, :] = evaluate(sample_lb, **kwargs)
            values_ub[i, :] = evaluate(sample_ub, **kwargs)
        baseline_2 = np.array(evaluate(sample, **kwargs))
        error = np.abs(baseline_2 - baseline_1)
        index, = np.where(error > 1e-6)
        error = error[index]
        relative_error = error / np.maximum.reduce([np.abs(baseline_1[index]), np.abs(baseline_2[index])])
        for i, idx in enumerate(index):
            if relative_error[i] > etol:
                print(RuntimeError(
                    f"inconsistent model; {indicators[idx]} has a value of "
                    f"{baseline_1[idx]} before evaluating sensitivity and "
                    f"{baseline_2[idx]} after"
                ))
        baseline = 0.5 * (baseline_1 + baseline_2)
        if array:
            return baseline, values_lb, values_ub
        else:
            indicator_index = var_columns(indicators)
            baseline = pd.Series(baseline, index=indicator_index)
            df_lb = pd.DataFrame(values_lb, 
                                index=var_columns(parameters),
                                columns=indicator_index)
            df_ub = df_lb.copy()
            df_ub[:] = values_ub
            return baseline, df_lb, df_ub
    
    def load_pickled_results(self, file=None, safe=True):
        table = self.table
        with open(file, "rb") as f:
            number, values, table_index, table_columns = pickle.load(f)
        if (table_index != table.index).any() or (table_columns != table.columns).any():
            if safe: raise ValueError('table layout does not match autoload file')
        del table_index, table_columns
        indicators = self._indicators
        table[var_indices(indicators)] = replace_nones(values, [np.nan] * len(indicators))
    
    def optimize(self, 
            loss, 
            parameters=None, 
            method=None, 
            convergence_model=None, 
            options=None,
        ):
        if parameters is None:
            parameters = self._optimized_parameters
        if method is None:
            method = self.default_optimizer
        else:
            method = method.lower()
        if options is None and method in self.default_optimizer_options: 
            options = self.default_optimizer_options[method]
        if isinstance(convergence_model, str):
            convergence_model = ConvergenceModel(
                system=self.system,
                predictors=parameters,
                model_type=convergence_model,
            )
        elif convergence_model is None:
            convergence_model = ConvergenceModel(
                system=self.system,
                predictors=parameters,
                model_type=self.default_convergence_model,
            )
        objective_function = self._objective_function
        args = (loss, parameters, convergence_model)
        bounds = np.array([p.bounds for p in parameters])
        if method == 'shgo':
            result = shgo(
                objective_function, bounds, args, options=options,
            )
        elif method == 'differential evolution':
            result = differential_evolution(
                objective_function, bounds, args, **options
            )
        else:
            raise ValueError(f'invalid optimization method {method!r}')
        return result, convergence_model
    
    def evaluate(self, notify=0, file=None, autosave=0, autoload=False,
                 convergence_model=None, **kwargs):
        """
        Evaluate indicators over the loaded samples and save values to `table`.
        
        Parameters
        ----------
        notify=0 : int, optional
            If 1 or greater, notify elapsed time after the given number of sample evaluations. 
        file : str, optional
            Name of file to save/load pickled evaluation results.
        autosave : int, optional
            If 1 or greater, save pickled evaluation results after the given number of sample evaluations.
        autoload : bool, optional
            Whether to load pickled evaluation results from file.
        convergence_model : ConvergencePredictionModel, optional
            A prediction model for accelerated system convergence. Defaults
            to no convergence model and the last solution
            as the initial guess for the next scenario. If a string is passed, 
            a ConvergenceModel will be created using that model type.
        kwargs : dict
            Any keyword arguments passed to :func:`biosteam.System.simulate`.
        
        Warning
        -------
        Any changes made to either the model or the samples will not be accounted
        for when autoloading and may lead to misleading results.
        
        """
        samples = self._samples
        if samples is None: raise RuntimeError('must load samples before evaluating')
        evaluate_sample = self._evaluate_sample
        table = self.table
        if isinstance(convergence_model, str):
            convergence_model = ConvergenceModel(
                system=self.system,
                predictors=self.parameters,
                model_type=convergence_model,
            )
        if notify:
            timer = Timer()
            timer.start()
            count = [0]
            def evaluate(sample, convergence_model, count=count, **kwargs):
                count[0] += 1
                values = evaluate_sample(sample, convergence_model, **kwargs)
                if not count[0] % notify:
                    print(f"{count} Elapsed time: {timer.elapsed_time:.0f} sec")
                return values
        else:
            evaluate = evaluate_sample
        N_samples, _ = samples.shape
        if autoload: 
            try:
                with open(file, "rb") as f:
                    number, values, table_index, table_columns = pickle.load(f)
                if (table_index != table.index).any() or (table_columns != table.columns).any():
                    raise ValueError('table layout does not match autoload file')
                del table_index, table_columns
                index = self._index[number:]
            except:
                number = 0
                index = self._index
                values = [None] * N_samples 
            else:
                if notify: count[0] = number
        else:
            number = 0
            index = self._index
            values = [None] * N_samples
        export = 'export_state_to' in kwargs
        layout = table.index, table.columns
        try:
            for number, i in enumerate(index, number + 1): 
                if export: kwargs['sample_id'] = i
                values[i] = evaluate(samples[i], convergence_model, **kwargs)
                if autosave and not number % autosave: 
                    obj = (number, values, *layout)
                    try:
                        with open(file, 'wb') as f: pickle.dump(obj, f)
                    except FileNotFoundError:
                        import os
                        head, tail = os.path.split(file)
                        os.mkdir(head)
                        with open(file, 'wb') as f: pickle.dump(obj, f)
        finally:
            table[var_indices(self._indicators)] = replace_nones(values, [np.nan] * len(self.indicators))
    
    def _reset_system(self):
        if self._system is None: return 
        self._system.empty_outlet_streams()
        self._system.reset_cache()
    
    def _failed_evaluation(self):
        self._reset_system()
        return [np.nan] * len(self.indicators)
    
    def indicators_at_baseline(self):
        """Return indicator values at baseline sample."""
        baseline = self.get_baseline_scenario()
        return self(baseline)
    metrics_at_baseline = indicators_at_baseline
    
    def evaluate_across_coordinate(self,
            name, f_coordinate, coordinate, *, 
            xlfile=None, notify=0, notify_coordinate=True,
            multi_coordinate=False, 
            simulation_independent_coordinate=False,
            f_evaluate=None
        ):
        """
        Evaluate across coordinate and save sample indicators.
        
        Parameters
        ----------
        name : str or tuple[str]
            Name of coordinate
        f_coordinate : function
            Should change state of system given the coordinate.
        coordinate : array
            Coordinate values.
        xlfile : str, optional
            Name of file to save. File must end with ".xlsx"
        rule : str, optional
            Sampling rule. Defaults to 'L'.
        notify_coordinate : bool, optional
            If True, notify elapsed time after each coordinate evaluation. Defaults to True.
        notify : int, optional
            Notify elapsed time after given number of scenario evaluations.
        f_evaluate : callable, optional
            Function to evaluate model. Defaults to evaluate method.
        
        """
        if (isinstance(f_coordinate, Parameter)
            and f_coordinate in self.parameters
            and f_coordinate.active):
            active = f_coordinate.active
            f_coordinate.active = False
            f_coordinate = f_coordinate.setter
            try:
                return self.evaluate_across_coordinate(
                    name, f_coordinate, coordinate,
                    xlfile=xlfile, notify=notify, notify_coordinate=notify_coordinate,
                    multi_coordinate=multi_coordinate,
                    simulation_independent_coordinate=simulation_independent_coordinate,
                    f_evaluate=f_evaluate
                )
            finally:
                f_coordinate.active = active
        N_points = len(coordinate)
        if simulation_independent_coordinate:
            if f_evaluate is not None: 
                raise ValueError(
                    'cannot pass `f_evaluate` if coordinate is '
                    'independent from simulation'
                )
            f_evaluate = self.evaluate
            samples = self._samples
            if samples is None: raise RuntimeError('must load samples before evaluating')
            evaluate_sample = self._evaluate_sample
            table = self.table
            if notify:
                timer = Timer()
                timer.start()
                count = [0]
                def evaluate(sample, count=count):
                    count[0] += 1
                    values = evaluate_sample(sample)
                    if not count[0] % notify:
                        print(f"{count} Elapsed time: {timer.elapsed_time:.0f} sec")
                    return values
            else:
                evaluate = evaluate_sample
            N_samples, _ = samples.shape
            number = 0
            index = self._index
            N_samples, _ = self.table.shape
            indicator_indices = var_indices(self.indicators)
            shape = (N_samples, N_points)
            indicator_data = {i: np.zeros(shape) for i in indicator_indices}
            for number, i in enumerate(index, number + 1): 
                evaluate(samples[i])
                for j, x in enumerate(coordinate):
                    f_coordinate(x)
                    for key, indicator in zip(indicator_data, self.indicators):
                        data = indicator_data[key]
                        try:
                            data[i, j] = indicator()
                        except:
                            data[i, j] = None
        else:
            if f_evaluate is None: f_evaluate = self.evaluate
            
            # Initialize timer
            if notify_coordinate:
                from biosteam.utils import Timer
                timer = Timer()
                timer.start()
                def evaluate(**kwargs):
                    f_evaluate(**kwargs)
                    print(f"[Coordinate {n}] Elapsed time: {timer.elapsed_time:.0f} sec")
            else:
                evaluate = f_evaluate
            indicator_data = None
            for n, x in enumerate(coordinate):
                f_coordinate(*x) if multi_coordinate else f_coordinate(x)
                evaluate(notify=notify)
                if indicator_data is None:
                    # Initialize data containers dynamically in case samples are loaded during evaluation
                    N_samples, _ = self.table.shape
                    indicator_indices = var_indices(self.indicators)
                    shape = (N_samples, N_points)
                    indicator_data = {i: np.zeros(shape) for i in indicator_indices}
                for indicator in indicator_data:
                    indicator_data[indicator][:, n] = self.table[indicator]
        
        if xlfile:
            if multi_coordinate:
                columns = pd.MultiIndex.from_tuples(coordinate,
                                                    names=name)
            else:
                columns = pd.Index(coordinate, name=name)
            
            # Save data to excel
            data = pd.DataFrame(data=np.zeros([N_samples, N_points]),
                                columns=columns)
            
            with pd.ExcelWriter(xlfile) as writer:
                for indicator in self.indicators:
                    data[:] = indicator_data[indicator.index]
                    data.to_excel(writer, sheet_name=indicator.short_description)
        return indicator_data
    
    def spearman(self, parameters=None, indicators=None):
        warn(DeprecationWarning('this method will be deprecated in biosteam 2.25; '
                                'use spearman_r instead'), stacklevel=2)
        return self.spearman_r(parameters, indicators)[0]
    
    def spearman_r(self, parameters=None, indicators=None, filter=None, **kwargs): # pragma: no cover
        """
        Return two DataFrame objects of Spearman's rho and p-values between indicators 
        and parameters.
        
        Parameters
        ----------
        parameters : Iterable[Parameter], optional
            Parameters to be correlated with indicators. Defaults to all parameters.
        indicators : Iterable[Indicator], optional 
            Indicators to be correlated with parameters. Defaults to all indicators.
        
        Other Parameters
        ----------------
        filter : Callable(x, y) -> x, y, or string, optional
            Function that accepts 1d arrays of x and y values and returns 
            filtered x and y values to correlate. May also
            be one of the following strings:
                
            * 'none': no filter
            
            * 'omit nan': all NaN values are ignored in correlation
            
            * 'propagate nan': NaN values return NaN correlation results
            
            * 'raise nan': NaN values will raise a ValueError
            
        **kwargs :
            Keyword arguments passed to :func:`scipy.stats.spearmanr`.
        
        See Also
        --------
        :func:`scipy.stats.spearmanr`
        
        """
        from scipy.stats import spearmanr
        return self._correlation(spearmanr, parameters, indicators, filter, kwargs)
    
    def pearson_r(self, parameters=None, indicators=None, filter=None, **kwargs):
        """
        Return two DataFrame objects of Pearson's rho and p-values between indicators 
        and parameters.
        
        Parameters
        ----------
        parameters : Iterable[Parameter], optional
            Parameters to be correlated with indicators. Defaults to all parameters.
        indicators : Iterable[Indicator], optional 
            Indicators to be correlated with parameters. Defaults to all indicators.
            
        Other Parameters
        ----------------
        filter : Callable(x, y) -> x, y, or string, optional
            Function that accepts 1d arrays of x and y values and returns 
            filtered x and y values to correlate. May also
            be one of the following strings:
                
            * 'none': no filter
            
            * 'omit nan': all NaN values are ignored in correlation
            
            * 'propagate nan': NaN values return NaN correlation results
            
            * 'raise nan': NaN values will raise a ValueError
            
        **kwargs :
            Keyword arguments passed to :func:`scipy.stats.pearsonr`.
        
        See Also
        --------
        :func:`scipy.stats.pearsonr`
        
        """
        from scipy.stats import pearsonr
        return self._correlation(pearsonr, parameters, indicators, filter, kwargs)
    
    def kendall_tau(self, parameters=None, indicators=None, filter=None, **kwargs):
        """
        Return two DataFrame objects of Kendall's tau and p-values between indicators 
        and parameters.
        
        Parameters
        ----------
        parameters : Iterable[Parameter], optional
            Parameters to be correlated with indicators. Defaults to all parameters.
        indicators : Iterable[Indicator], optional 
            Indicators to be correlated with parameters. Defaults to all indicators.
            
        Other Parameters
        ----------------
        filter : Callable(x, y) -> x, y, or string, optional
            Function that accepts 1d arrays of x and y values and returns 
            filtered x and y values to correlate. May also
            be one of the following strings:
            
            * 'none': no filter
            
            * 'omit nan': all NaN values are ignored in correlation
            
            * 'propagate nan': NaN values return NaN correlation results
            
            * 'raise nan': NaN values will raise a ValueError
        
        **kwargs :
            Keyword arguments passed to :func:`scipy.stats.kendalltau`.
        
        See Also
        --------
        :func:`scipy.stats.kendalltau`
        
        """
        from scipy.stats import kendalltau
        return self._correlation(kendalltau, parameters, indicators, filter, kwargs)
    
    def kolmogorov_smirnov_d(self, parameters=None, indicators=None, thresholds=[],
                             filter=None, **kwargs):
        """
        Return two DataFrame objects of Kolmogorov–Smirnov's D and p-values
        with the given thresholds for the indicators.
        
        For one particular parameter, all of the sampled values will be divided into two sets,
        one where the resulting indicator value is smaller than or equal to the threshold,
        and the other where the resulting value is larger than the threshold.
        
        Kolmogorov–Smirnov test will then be performed for these two sets of values
        for the particular parameter.
        
        Parameters
        ----------
        parameters : Iterable[Parameter], optional
            Parameters to be correlated with indicators. Defaults to all parameters.
        indicators : Iterable[Indicator], optional 
            Indicators to be correlated with parameters. Defaults to all indicators.
        thresholds : Iterable[float]
            The thresholds for separating parameters into sets.
            
        Other Parameters
        ----------------
        filter : Callable(x, y) -> x, y, or string, optional
            Function that accepts 1d arrays of x and y values and returns 
            filtered x and y values to correlate. May also
            be one of the following strings:
                
            * 'none': no filter
            
            * 'omit nan': all NaN values are ignored in correlation
            
            * 'propagate nan': NaN values return NaN correlation results
            
            * 'raise nan': NaN values will raise a ValueError
            
        **kwargs :
            Keyword arguments passed to :func:`scipy.stats.kstest`.
        
        See Also
        --------
        :func:`scipy.stats.kstest`
        
        """
        from scipy.stats import kstest
        indicators = indicators or self.indicators
        if len(thresholds) != len(indicators):
            raise ValueError(f'The number of indicators {len(indicators)} must match '
                             f'the number of thresholds ({len(thresholds)}).')
        kwargs['thresholds'] = thresholds
        return self._correlation(kstest, parameters, indicators, filter, kwargs)
    
    def _correlation(self, f, parameters, indicators, filter, kwargs):
        """
        Return two DataFrame objects of statistics and p-values between indicators 
        and parameters.
        
        Parameters
        ----------
        f : Callable
            Function with signature f(x, y) -> stat, p
        parameters : Iterable[Parameter], optional
            Parameters to be correlated with indicators. Defaults to all parameters.
        indicators : Iterable[Indicator], optional 
            Indicators to be correlated with parameters. Defaults to all indicators.
        filter : Callable or string, optional
            Function that accepts 1d arrays of x and y values and returns 
            filtered x and y values to correlate. May also
            be one of the following strings:
                
            * 'none': no filter
            
            * 'omit nan': all NaN values are ignored in correlation
            
            * 'propagate nan': NaN values return NaN correlation results
            
            * 'raise nan': NaN values will raise a ValueError
            
        kwargs : dict
            Keyword arguments passed to `f`.
            
        """
        if not parameters: parameters = self._parameters
        table = self.table
        values = table.values.transpose()
        index = table.columns.get_loc
        parameter_indices = var_indices(parameters)
        parameter_data = [values[index(i)] for i in parameter_indices]
        indicator_indices = var_indices(indicators or self.indicators)
        indicator_data = [values[index(i)] for i in indicator_indices]
                
        if not filter: filter = 'propagate nan'
        if isinstance(filter, str):
            name = filter.lower()
            if name == 'omit nan':
                def corr(x, y):
                    index = ~(np.isnan(x) | np.isnan(y))
                    return f(x[index], y[index], **kwargs)
            elif name == 'propagate nan':
                def corr(x, y):
                    if np.isnan(x).any() or np.isnan(y).any(): 
                        NaN = float('NaN')
                        return NaN, NaN
                    else:
                        return f(x, y, **kwargs)
            elif name == 'raise nan':
                def corr(x, y):
                    if np.isnan(x).any() or np.isnan(y).any():
                        raise ValueError('table entries contain NaN values')
                    return f(x, y)
            elif name == 'none':
                corr = lambda x, y: f(x, y, **kwargs)
            else: #: pragma: no cover
                raise ValueError(
                    f"invalid filter '{filter}'; valid filter names are: "
                    "'omit nan', 'propagate nan', 'raise nan', and 'none'"
                )
        elif callable(filter): 
            corr = lambda x, y: f(*filter(x, y), **kwargs)
        else: #: pragma: no cover
            raise TypeError("filter must be either a string or a callable; "
                            "not a '{type(filter).__name__}' object")

        if 'thresholds' not in kwargs:
            data = np.array([[corr(p.astype(float), m.astype(float)) for m in indicator_data] for p in parameter_data])
        else: # KS test
            thresholds = kwargs.pop('thresholds')
            data = np.array(
                [
                    [corr(parameter_data[n_p][indicator_data[n_m]<=thresholds[n_m]],
                          parameter_data[n_p][indicator_data[n_m]>thresholds[n_m]]) \
                     for n_m in range(len(indicator_data))] \
                for n_p in range(len(parameter_data))])
        
        index = indices_to_multiindex(parameter_indices, ('Element', 'Parameter'))
        columns = indices_to_multiindex(indicator_indices, ('Element', 'Indicator'))        
        return [pd.DataFrame(i, index=index, columns=columns) for i in (data[..., 0], data[..., 1])]
        
    def create_fitted_model(self, parameters, indicators): # pragma: no cover
        from pipeml import FittedModel
        Xdf = self.table[[i.index for i in parameters]]
        ydf_index = indicators.index if isinstance(indicators, Indicator) else [i.index for i in indicators]
        ydf = self.table[ydf_index]
        return FittedModel.from_dfs(Xdf, ydf)
    
    def __call__(self, sample, **kwargs):
        """Return pandas Series of indicator values at given sample."""
        self._update_state(np.asarray(sample, dtype=float), **kwargs)
        return pd.Series({i.index: i() for i in self._indicators})
    
    def _repr(self, m):
        clsname = type(self).__name__
        newline = "\n" + " "*(len(clsname)+2)
        return f'{clsname}: {newline.join([i.describe() for i in self.indicators])}'
    
    def __repr__(self):
        return f'<{type(self).__name__}: {len(self.parameters)}-parameters, {len(self.indicators)}-indicators>'
    
    def _info(self, p, m):
        info = f'{type(self).__name__}:'
        parameters = self._parameters
        if parameters: 
            p_max = len(parameters)
            if p is None: p = p_max
            ptitle = 'parameters: '
            info += '\n' + ptitle
            newline = "\n" + " "*(len(ptitle))
            info += newline.join([
                parameters[i].describe(
                    distribution=False, bounds=False
                ) for i in range(p)
            ])
            if p < p_max: info += newline + "..." 
        else:
            info += '\nparameters: None'
        indicators = self._indicators
        if indicators: 
            m_max = len(indicators)
            if m is None: m = m_max
            mtitle = 'indicators: '
            info += '\n' + mtitle
            newline = "\n" + " "*(len(mtitle))
            info += newline.join([
                indicators[i].describe() for i in range(m)
            ])
            if m < m_max: info += newline + "..." 
        else:
            info += '\nindicators: None'
        return info
    
    def show(self, p=None, m=None):
        """Return information on p-parameters and m-indicators."""
        print(self._info(p, m))
    _ipython_display_ = show
    
EasyInputModel = Model
Model.load_parameter_distributions = Model.parameters_from_df



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from thermosteam import utils
from . import (
    patches,
    timer,
    piping,
    stream_link_options,
    functors,
    scope,
)
__all__ = (
    'colors',
    'patches', 
    'functors',
    *utils.__all__,
    *patches.__all__, 
    *timer.__all__, 
    *piping.__all__, 
    *stream_link_options.__all__,
    *functors.__all__,
    *scope.__all__,
)
from thermosteam.utils import *
from .patches import *
from .timer import *
from .piping import *
from .stream_link_options import *
from .functors import *
from .scope import *

del utils


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import pytest
import biosteam as bst
from numpy.testing import assert_allclose

# def test_fit():
#     import biosteam as bst
#     import numpy as np
#     Ce = np.array([
#         0.06013, 0.07133, 0.07133, 0.31769, 0.31769, 0.31769,
#         0.63124, 0.64244, 0.63124, 2.03102, 2.03102, 2.03102,
#         4.16988, 4.16988, 4.16988, 5.00974, 4.99854, 4.99854
#     ]) / 1000
#     qe = np.array([
#         1.24, 1.24, 1.24, 1.87, 1.87, 1.87, 2.85, 2.84, 2.85,
#         4.48, 4.48, 4.48, 5.5, 5.5, 5.5, 6.53, 6.56, 6.56
#     ])
#     t = np.array([0.0, 31.0, 60.0, 90.0, 120.0, 150.0]) / 60
#     C = np.array([7.0, 2.45, 1.03, 0.43, 0.26, 0.13])
#     volume = 0.075
#     adsorbent = 1
#     bst.AdsorptionColumn.plot_isotherm_and_mass_transfer_coefficient_fit(
#         Ce, qe, t, C, volume, adsorbent,
#     )
    
# def test_adsorption():
#     import biosteam as bst
#     chemicals = bst.Chemicals([
#         'Xylene', 
#         bst.Chemical('Ink', search_db=False, default=True, phase='l'),
#         bst.Chemical('ActivatedCarbon', search_db=False, default=True, phase='l')
#     ])
#     bst.settings.set_thermo(chemicals)
#     chemicals.compile()
#     chemicals.define_group('Solvent', ['Xylene'])
#     feed = bst.Stream(ID='feed', phase='l', T=298, P=1.01e+06,
#                       Xylene=1000, Ink=0.001, units='kg/hr')
#     # feed.T = 450
#     # Adsorption equilibrium and rate data from Dr. Tianwei Yan from Prof. George Huber's group.
#     A1 = bst.AdsorptionColumn(
#         'A1', ins=feed, outs='effluent',
#         cycle_time=1000,
#         superficial_velocity=4,
#         isotherm_model='Freundlich',
#         isotherm_args = (48.4, 2.58), 
#         k = 0.0093, # [1 / hr]
#         # Equilibrium constant [L/g] for Langmuir isotherm (multiply by 1000 from L/mg)
#         # Maximum equilibrium loading [mg/g] for Langmuir isotherm
#         # isotherm_model='Langmuir',
#         # isotherm_args = (1.27e3, 6.99), 
#         # k = 0.13, #[1 / hr]
#         void_fraction=1 - 0.38 / 0.8,  # Solid by wt [%]
#         C_final_scaled=0.05,
#         adsorbate='Ink',
#     )
#     A1.simulate()
#     assert_allclose(A1.outs[0].mol, [9.419, 0., 0.], rtol=1e-3)
#     assert 1e-3 < A1.CL_scaled[-1, -1] < 2 * A1.C_final_scaled # Shape of MTZ is not exact
#     assert_allclose(A1.CL_scaled[-1, -1], 0.0842941306073549, rtol=1e-3)
#     assert_allclose(A1.LES, 2.504118270726701)
#     assert_allclose(A1.q0, 3.6870766727528457)
#     assert_allclose(A1.design_results['Diameter'], 1.9764225277140537)
#     assert_allclose(A1.get_design_result('Length', 'm'), A1.LES + A1.LUB)

# def test_adsorption_bed_pressure_drop():
#     f = lambda L, u: bst.unuts.adsorption.adsorption_bed_pressure_drop(u=u/3600, L=L)
#     X, Y, Z = bst.plots.generate_contour_data(f, [1, 10], [4, 14], n=10)
    



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
__all__ = ('Feature', 'MockFeature', 'Variable', 'MockVariable')
from ._name import element_name
from ..utils import format_title
from thermosteam import units_of_measure

class Feature:
    """
    Abstract class for a feature in BioSTEAM.
    
    Attributes
    ----------
    name : str
        Name of feature.
    units : str
        Units of measure.
    element : object
        Element corresponding to feature.
        
    """
    __slots__ = ('name', 'units', 'element')
    include_units_in_index = True
    
    def __init__(self, name, units, element):
        self.name = format_title(name)
        self.units = units
        self.element = element
    
    def mockup(self):
        return MockFeature(self.name, self.units, self.element)
    
    @classmethod
    def check_index_unique(cls, feature, features):
        key = (feature.element_name, feature.name)
        keys = {(i.element_name, i.name) for i in features}
        if key in keys:
            raise ValueError(
                     "each feature must have a unique element and name; "
                    f"feature with element {repr(feature.element)} "
                    f"and name {repr(feature.name)} already present"
                )
    
    @classmethod
    def check_indices_unique(cls, features):
        keys = set()
        for i in features:
            key = (i.element, i.name)
            if key in keys:
                kind = cls.__name__.lower()
                raise ValueError(
                        f"each {kind} must have a unique element and name; "
                        f"more than one {kind} with element {repr(i.element)} "
                        f"and name {repr(i.name)} are present"
                    )
            keys.add(key)
    
    @property
    def element_name(self):
        return element_name(self.element)
    
    @property
    def name_with_units(self):
        units = self.units
        name = self.name
        if units: name += f" [{units}]"
        return name
    
    @property
    def index(self):
        name = self.name
        if self.include_units_in_index:
            units = self.units
            if units: name += f" [{units}]"
        return (self.element_name, name)
    
    @property
    def short_description(self):
        element, name = self.index
        name, *_ = name.split(' [')
        if element not in name:
            name = ' '.join([element, name])
        if len(name) > 31:
            words = name.split(' ')
            words = [(i[:4]+'.' if len(i) > 5 else i) for i in words]
            name = ' '.join(words)
        name = name.strip(' ')
        if len(name) > 31: name = name[:31]
        return name
    
    def label(self, element=True, element_dlim='\n', format_units=True):
        return self.describe(
            element_dlim, 
            format_units,
            False, 
            False, 
            element=element,
        )
    
    def describe(self, 
            element_dlim=' - ',
            format_units=False,
            distribution=True, 
            bounds=True, 
            number_format='.3g', 
            element=True,
        ) -> str:
        """Return description of feature."""
        name = self.name
        if element and self.element:
            name = self.element_name + element_dlim + name
        units = self.units
        if format_units and units:
            units = units_of_measure.format_units(units)
        description = name
        units_added = False
        if distribution:
            if getattr(self, 'distribution', None):
                dist_name = type(self.distribution).__name__
                distribution_values = self.distribution._repr.values()
                distribution = ', '.join([format(j, number_format)
                                          for j in distribution_values])
                if units:
                    description += f' ({dist_name}; {distribution} {str(units)})'
                else:
                    description += f' ({dist_name}; {distribution})'
                units_added = True
        elif bounds:
            baseline = getattr(self, 'baseline', None)
            bounds = getattr(self, 'bounds', None)
            if bounds and baseline:
                lb, ub = bounds
                values = ', '.join([format(i, number_format)
                                    for i in (lb, baseline, ub)])
                if units:
                    description += f' ({values} {str(units)})'
                else:
                    description += f' ({values})'
                units_added = True
        if description:
            first_letter = description[0]
            if first_letter.islower(): 
                description = first_letter.upper() + description[1:]
        if units and not units_added: 
            description += ' [' + str(units) + ']'
        return description
    
    def __repr__(self):
        units = f" ({self.units})" if self.units else ""
        element = f" [{self.element_name}]" if self.element else ""
        return f'<{type(self).__name__}:{element} {self.name}{units}>'
    
    def show(self):
        print(self._info())
     

class MockFeature(Feature):
    __slots__ = ()
    
    def __init__(self, name, units, element):
        super().__init__(name, units, element_name(element))
    
    def __repr__(self):
        return f"{type(self).__name__}('{self.name}', '{self.units}', '{self.element}')"

Variable = Feature
MockVariable = MockFeature 


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import pytest
import biosteam as bst
from numpy.testing import assert_allclose
from thermo import SRK, PR


def test_isentropic_helium_turbine():
    bst.settings.set_thermo(["He"], cache=True)
    eta = 1.0
    feed = bst.Stream(He=1, T=-160 + 273.15, P=20e5, phase='g')
    K = bst.units.IsentropicTurbine(ins=feed, P=5e5, eta=eta)
    K.simulate()
    # check outlet state
    out = K.outs[0]
    assert_allclose(
        [out.vapor_fraction, out.liquid_fraction, out.T, out.P],
        [1.0, 0.0, -208.16+273.15, 5e5],
        rtol=1e-4,
    )
    # check compressor design
    ideal_power = -0.278
    eta_motor = K.efficiency
    expected_power = ideal_power * eta * eta_motor
    actual_power = K.power_utility.rate
    assert_allclose(
        [actual_power, K.design_results['Ideal power'],
           K.design_results['Ideal duty'], K.design_results['Turbines in parallel']],
        [expected_power, ideal_power, 0, 1],
        rtol=1e-3,
    )
    pass

def test_isentropic_nitrogen_liquefaction():
    thermo = bst.Thermo([bst.Chemical('N2')])
    thermo.mixture.include_excess_energies = True
    bst.settings.set_thermo(thermo, cache=True)
    eta = 1.0
    feed = bst.Stream(N2=1, T=-108 + 273.15, P=200e5, phase='g')
    K = bst.units.IsentropicTurbine(ins=feed, P=1e5, eta=eta, vle=True)
    for i in range(3): K.simulate()
    # check outlet state
    out = K.outs[0]
    assert_allclose(
        [out.vapor_fraction, out.liquid_fraction, out.T, out.P],
        [5.284013e-01, 4.715987e-01, 7.724350e+01, 100000.0],
        rtol=5e-3,
    )
    # check compressor design
    ideal_power = -0.597828
    eta_motor = K.efficiency
    expected_power = ideal_power * eta * eta_motor
    actual_power = K.power_utility.rate
    assert_allclose(
        [actual_power, K.design_results['Ideal power'],
           K.design_results['Ideal duty'], K.design_results['Turbines in parallel']],
        [expected_power, ideal_power, 0, 1],
        rtol=5e-3,
    )

def test_turbine_invalid_pressure_warning():
    bst.settings.set_thermo(["N2"], cache=True)
    feed = bst.Stream(N2=1, T=298, P=1e5, phase='g')
    K = bst.units.IsentropicTurbine(ins=feed, P=2e5, eta=1, vle=False)
    with pytest.warns(RuntimeWarning):
        K.simulate()


if __name__ == '__main__':
    test_isentropic_helium_turbine()
    test_isentropic_nitrogen_liquefaction()
    test_turbine_invalid_pressure_warning()



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
.. contents:: :local:

.. autoclass:: biosteam.units.vacuum_system.VacuumSystem

"""
from . import design_tools as design
from .auxiliary import Auxiliary
from typing import Optional, Union
import biosteam as bst
from .._unit import Unit

__all__ = ('VacuumSystem',)
       
class VacuumSystem(Auxiliary): 
    """
    Create an auxiliary vacuum system for a unit operation. 
    
    Parameters
    ----------
    unit : 
        Main unit operation containing the auxiliary vacuum system. Missing parameters
        are estimated through this unit operation, if given.
    F_mass : 
        Vapor mass flow rate entering vacuum system from vessel [kg/hr] (not including inleakage).
        Defaults to the vapor volumetric flow rate exiting `unit`.
    F_vol : 
        Vapor volumetric flow rate entering vacuum system from vessel [m3/hr] (not including inleakage).
        Defaults to the vapor mass flow rate exiting `unit`.
    P_suction : 
        Suction pressure [Pa]. Defaults to `unit.P` or the minimum outlet pressure.
    vessel_volume : 
        Vacuum volume [m3]. Defaults to `unit.design_results['Total volume']`.
    vacuum_system_preference : 
        Name(s) of preferred vacuum systems. Valid names include 'Liquid-ring pump', 
        'Steam-jet ejector', and 'Dry-vacuum pump'.
    
    Notes
    -----
    The vacuum system is sized/costed based on the vapor flow rate through the 
    vacuum system, which includes the inleakage into the vessel through fixtures 
    and cracks. The inleakage is a function of the suction pressure and the 
    total vessel volume as in [1]_. 
    
    BioSTEAM's CSTR, Flash, and Distillation columns automatically include a 
    vacuum system when needed.
    
    Examples
    --------
    Create vessel with a vacuum system:
        
    >>> import biosteam as bst
    >>> class VacuumVessel(bst.Unit):
    ...     auxiliary_unit_names = ('vacuum_system',) # Mark attributes as auxiliary
    ...     _units = {'Total volume': 'm3'} # This is needed for the vacuum system
    ...     P = 1000 # Pa
    ...     tau = 4 # hr
    ...
    ...     def _run(self):
    ...         self.outs[0].P = 1000 # Pa
    ...     
    ...     def _design(self):
    ...         self.design_results['Total volume'] = self.feed.F_vol * self.tau
    ...         self.vacuum_system = bst.VacuumSystem(self)
    ...
    >>> bst.settings.set_thermo(['Water'])
    >>> feed = bst.Stream('feed', Water=1e6)
    >>> V1 = VacuumVessel('V1', ins=feed)
    >>> V1.simulate()
    >>> V1.results()
    Vacuum vessel                                                Units        V1
    Medium pressure steam Duty                                   kJ/hr  1.04e+07
                          Flow                                 kmol/hr       288
                          Cost                                  USD/hr      79.3
    Cooling water         Duty                                   kJ/hr -9.37e+06
                          Flow                                 kmol/hr   6.4e+03
                          Cost                                  USD/hr      3.12
    Design                Total volume                              m3  7.23e+04
    Purchase cost         Vacuum system - Steam-jet ejecto...      USD  8.18e+04
    Total purchase cost                                            USD  8.18e+04
    Utility cost                                                USD/hr      82.4
    
    For simplicity, this example does not include the cost of the vessel, but 
    vessel costs should be included for techno-economic analysis.
    
    References
    ----------
    .. [1] Seider, W. D.; Lewin, D. R.; Seader, J. D.; Widagdo, S.; Gani, R.;
        Ng, M. K. Cost Accounting and Capital Cost Estimation.
        In Product and Process Design Principles; Wiley, 2017; pp 426–485.
    
    """
    __slots__ = (
        'power_utility',
        'heat_utilities', 
        'baseline_purchase_costs',
        'purchase_costs',
        'installed_costs',
        'F_M', 'F_D', 'F_P', 'F_BM',
        'owner',
    )
    
    def __init__(self,
            unit: Optional[Unit]=None, 
            vacuum_system_preference: Optional[str]=None, 
            F_mass: Optional[float]=None,
            F_vol: Optional[float]=None, 
            P_suction: Optional[float]=None, 
            vessel_volume: Optional[Union[float, list[float]]]=None
        ):
        super().__init__()
        if unit:
            # Deduce arguments if unit given
            if F_mass is None or F_vol is None:
                # If vapor is condensed, assume vacuum system is after condenser
                vents = [i for i in unit.outs if 'g' in i.phase]
                F_mass = 0.
                F_vol = 0.
                for vapor in vents:
                    hx = vapor.sink
                    if isinstance(hx, bst.HX):
                        index = hx.ins.index(vapor)
                        vapor = hx.outs[index]
                        if 'g' not in vapor.phase: continue
                    if isinstance(vapor, bst.MultiStream):
                        F_mass += vapor['g'].F_mass
                        F_vol += vapor['g'].F_vol
                    else:
                        F_mass += vapor.F_mass
                        F_vol += vapor.F_vol
            if P_suction is None:
                P_suction = getattr(unit, 'P', None) or getattr(unit, 'P_suction', None)
                if P_suction is None: P_suction = min([i.P for i in unit.outs])
            if vessel_volume is None:
                if 'Total volume' in unit.design_results:
                    vessel_volume = unit.get_design_result('Total volume', 'm3')
                elif 'Volume' in unit.design_results:
                    vessel_volume = unit.get_design_result('Volume', 'm3')
                else:
                    raise ValueError("'Total volume' was not found in design results; "
                                     "'vessel_volume' parameter could not be deduced")
        else:
            # In case user does not supply all arguments
            params = [('F_mass', F_mass), 
                      ('F_vol', F_vol),
                      ('P_suction', P_suction), 
                      ('vessel_volume', vessel_volume)]
            for name, value in params:
                if value is None: 
                    raise ValueError(
                        f"missing argument '{name}'; cannot deduce '{name}'"
                         "when no unit is specified"
                    )
        capex = self.baseline_purchase_costs # Assume all costs the same
        vacuum_results = design.compute_vacuum_system_power_and_cost(
            F_mass, F_vol, P_suction, vessel_volume, vacuum_system_preference
        )
        name = vacuum_results['Name']
        capex[name] = vacuum_results['Cost']
        heating_agent = vacuum_results['Heating agent']
        if heating_agent: # Steam turbine
            vacuum_steam = self.create_heat_utility()
            vacuum_steam.set_utility_by_flow_rate(heating_agent, vacuum_results['Steam flow rate'])
            if vacuum_results['Condenser']: 
                self.add_heat_utility(-vacuum_steam.unit_duty, 373.15) # Vacuum cooling water
        self.add_power_utility(vacuum_results['Work'])
